"""
Tests for ``nefelibata.builders.html``.
"""
# pylint: disable=invalid-name
from pathlib import Path
from typing import Any, Dict

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture

from nefelibata import __version__
from nefelibata.builders.html import HTMLBuilder
from nefelibata.config import Config, SocialModel
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_builder_setup(root: Path, config: Config) -> None:
    """
    Test that templates and build directory are created.
    """
    template_directory = root / "templates/builders/html/minimal/src"
    build_directory = root / "build/html"
    assert not template_directory.exists()
    assert not build_directory.exists()

    builder = HTMLBuilder(root, config, "https://example.com/")
    builder.setup()

    last_update: Dict[Any, Any] = {}
    assert template_directory.exists()
    for file in ("index.html", "atom.xml", "post.html"):
        assert (template_directory / file).exists()
        last_update[file] = (template_directory / file).stat().st_mtime
    assert build_directory.exists()
    last_update[build_directory] = build_directory.stat().st_mtime

    # on the second call the files and directories should be unmodified
    HTMLBuilder(root, config, "https://example.com/")
    for file in ("index.html", "atom.xml", "post.html"):
        assert (template_directory / file).stat().st_mtime == last_update[file]
    assert build_directory.stat().st_mtime == last_update[build_directory]


@pytest.mark.asyncio
async def test_builder_post(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test ``process_post``.
    """
    _logger = mocker.patch("nefelibata.builders.base._logger")

    builder = HTMLBuilder(root, config, "https://example.com/")
    builder.setup()

    # call twice to check that it won't fail (shouldn't overwrite)
    builder.setup()

    with freeze_time("2021-01-02T00:00:00Z"):
        await builder.process_post(post)

    post_path = root / "build/html/first/index.html"

    # test that file was created
    assert post_path.exists()
    _logger.info.assert_called_with("Creating %s post", "HTML")
    with open(post_path, encoding="utf-8") as input_:
        content = input_.read()
    assert (
        content
        == f"""<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <meta name="generator" content="Nefelibata {__version__}">
    <meta name="robots" content="index, follow">
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="description" content="Musings about the path and other things">

    <title>道&amp;c.: This is your first post</title>

    <link rel="canonical" href="https://example.com/" />
    <link rel="alternate" type="application/atom+xml" href="https://example.com/atom.xml" />
    <link rel="webmention" href="https://webmention.io/example.com/webmention" />
    <link rel="me" href="mailto:roberto@dealmeida.net" />

    <link rel="stylesheet" href="https://example.com/css/awsm_theme_big-stone.min.css" media="(prefers-color-scheme: dark)">
    <link rel="stylesheet" href="https://example.com/css/awsm_theme_pearl-lusta.min.css" media="(prefers-color-scheme: no-preference), (prefers-color-scheme: light)">
    <link rel="stylesheet" href="https://example.com/css/custom.css">
  </head>
  <body>
    <main>
      <nav>
        <ul>
          <li><a href="https://example.com/">Main index</a></li>
        </ul>
      </nav>

      <article class="h-entry">
        <h1 class="p-name">This is your first post</h1>

        <p>Published by <a class="p-author h-card" href="https://taoetc.org/">Beto Dealmeida</a>
        on <time class="dt-published" datetime="2020-12-31T16:00:00-08:00">2020-12-31 16:00:00-08:00</time></p>

        <p class="p-summary"><small>Hello, world!</small></p>

        <div class="e-content">
          <h1>Welcome</h1>
<p>This is your first post. It should be written using Markdown.</p>
<p>Read more about <a href="https://nefelibata.readthedocs.io/">Nefelibata</a>.</p>

        </div>


        <h2>Tags</h2>

        <ul>
          <li><a class="p-category" href="https://example.com/tags/blog.html">blog</a></li>
          <li><a class="p-category" href="https://example.com/tags/welcome.html">welcome</a></li>
        </ul>

        <h2>Categories</h2>

        <ul>
          <li><a class="p-category" href="https://example.com/categories/stem.html">STEM</a></li>
        </ul>

      </article>
    </main>
  </body>
</html>"""
    )

    # call again, test that file is up-to-date
    _logger.reset_mock()
    last_update = post_path.stat().st_mtime
    with freeze_time("2021-01-03T00:00:00Z"):
        await builder.process_post(post)
    assert post_path.stat().st_mtime == last_update
    _logger.debug.assert_called_with("Post %s is up-to-date, nothing to do", post_path)

    # call again, forcing a rebuild
    _logger.reset_mock()
    last_update = post_path.stat().st_mtime
    with freeze_time("2021-01-04T00:00:00Z"):
        await builder.process_post(post, force=True)
    assert post_path.stat().st_mtime > last_update
    _logger.info.assert_called_with("Creating %s post", "HTML")


@pytest.mark.asyncio
async def test_builder_site(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test ``process_site``.
    """
    _logger = mocker.patch("nefelibata.builders.base._logger")
    mocker.patch("nefelibata.builders.base.get_posts", return_value=[post])

    config.social = [SocialModel(title="Mastodon", url="https://2c.taoetc.org/@beto")]
    builder = HTMLBuilder(root, config, "https://example.com/")
    builder.setup()

    with freeze_time("2021-01-02T00:00:00Z"):
        await builder.process_site()

    assets_directory = root / "build/html"
    assets = (
        "index.html",
        "atom.xml",
        "tags/welcome.html",
        "tags/blog.html",
        "categories/stem.html",
    )

    # test that files were created
    last_update: Dict[Any, Any] = {}
    for asset in assets:
        assert (assets_directory / asset).exists()
        last_update[asset] = (assets_directory / asset).stat().st_mtime
    _logger.info.assert_has_calls(
        [
            mocker.call("Creating %s", Path("/path/to/blog/build/html/index.html")),
            mocker.call("Creating %s", Path("/path/to/blog/build/html/atom.xml")),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/html/tags/blog.html"),
            ),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/html/tags/welcome.html"),
            ),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/html/categories/stem.html"),
            ),
        ],
    )
    with open(assets_directory / "index.html", encoding="utf-8") as input_:
        content = input_.read()
    assert (
        content
        == f"""<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <meta name="generator" content="Nefelibata {__version__}">
    <meta name="robots" content="index, follow">
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="description" content="Musings about the path and other things">

    <title>道&amp;c.: Musings about the path and other things</title>

    <link rel="canonical" href="https://example.com/" />
    <link rel="alternate" type="application/atom+xml" href="https://example.com/atom.xml" />
    <link rel="webmention" href="https://webmention.io/example.com/webmention" />
    <link rel="me" href="mailto:roberto@dealmeida.net" />

    <link rel="stylesheet" href="https://example.com/css/awsm_theme_big-stone.min.css" media="(prefers-color-scheme: dark)">
    <link rel="stylesheet" href="https://example.com/css/awsm_theme_pearl-lusta.min.css" media="(prefers-color-scheme: no-preference), (prefers-color-scheme: light)">
    <link rel="stylesheet" href="https://example.com/css/custom.css">
  </head>
  <body>
    <main>
      <h1>道&amp;c.: Musings about the path and other things</h1>

      <h2 class="h-feed">Posts</h2>

      <ul>
        <li class="h-entry"><a class="p-name" href="https://example.com/first/index.html">This is your first post</a> </li>
      </ul>

      <h2>Links</h2>

      <ul>
        <li><a href="https://2c.taoetc.org/@beto" rel="me">Mastodon</a></li>
      </ul>

      <footer>
        <p>Crafted with ❤️ using <a href="https://nefelibata.readthedocs.io/">Nefelibata</a>.</p>
      </footer>
    </main>
  </body>
</html>"""
    )
    with open(assets_directory / "atom.xml", encoding="utf-8") as input_:
        content = input_.read()
    assert (
        content
        == """<?xml version="1.0" encoding="utf-8"?>
<feed xmlns="http://www.w3.org/2005/Atom">
    <title>道&amp;c.</title>
    <link rel="self" type="application/atom+xml" href="https://example.com/atom.xml" />
    <id>https://example.com/</id>
    <updated>2020-12-31T16:04:00Z</updated>
<entry>
    <title>This is your first post</title>
    <link rel="alternate" type="text/html" href="https://example.com/first/index.html" />
    <id>tag:example.com,2020-12-31:first/index.html</id>
    <updated>2020-12-31T16:04:00Z</updated>
    <summary type="html">
        Hello, world!
        &lt;p&gt;&lt;a href="https://example.com/first/index.html"&gt;Permalink&lt;/p&gt;
    </summary>
    <author>
        <name>Beto Dealmeida</name>
        <email>roberto@dealmeida.net</email>
    </author>
    <category term="stem" />
    <content type="html">
        &lt;h1&gt;Welcome&lt;/h1&gt;
&lt;p&gt;This is your first post. It should be written using Markdown.&lt;/p&gt;
&lt;p&gt;Read more about &lt;a href=&#34;https://nefelibata.readthedocs.io/&#34;&gt;Nefelibata&lt;/a&gt;.&lt;/p&gt;

        &lt;p&gt;&lt;a href="https://example.com/first/index.html"&gt;Permalink&lt;/p&gt;
    </content>
</entry>
</feed>"""
    )

    # call again, test that file is up-to-date
    _logger.reset_mock()
    with freeze_time("2021-01-03T00:00:00Z"):
        await builder.process_site()
    for asset in assets:
        assert (assets_directory / asset).stat().st_mtime == last_update[asset]
    _logger.debug.assert_has_calls(
        [
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/html/index.html"),
            ),
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/html/atom.xml"),
            ),
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/html/tags/blog.html"),
            ),
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/html/tags/welcome.html"),
            ),
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/html/categories/stem.html"),
            ),
        ],
    )

    # call again, forcing a rebuild
    _logger.reset_mock()
    with freeze_time("2021-01-04T00:00:00Z"):
        await builder.process_site(force=True)
    for asset in assets:
        assert (assets_directory / asset).stat().st_mtime > last_update[asset]
    _logger.info.assert_has_calls(
        [
            mocker.call("Creating %s", Path("/path/to/blog/build/html/index.html")),
            mocker.call("Creating %s", Path("/path/to/blog/build/html/atom.xml")),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/html/tags/blog.html"),
            ),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/html/tags/welcome.html"),
            ),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/html/categories/stem.html"),
            ),
        ],
    )



"""
A CAPCOM (gemini://gemini.circumlunar.space/capcom/) announcer.
"""

import logging
from datetime import timedelta

from nefelibata.announcers.gemlog import GemlogAnnouncer

_logger = logging.getLogger(__name__)


class CAPCOMAnnouncer(GemlogAnnouncer):

    """
    A CAPCOM (gemini://gemini.circumlunar.space/capcom/) announcer.

    This requires the Gemini builder, and assumes that the Gemlog feed is
    at ``{builder.home}/feed{builder.extension}``.
    """

    name = "CAPCOM"
    url = "gemini://gemini.circumlunar.space/capcom/"
    submit_url = "gemini://gemini.circumlunar.space/capcom/submit?"
    grace_seconds = timedelta(days=365).total_seconds()
    logger = _logger



"""
Publish the blog.
"""
import asyncio
import logging
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, Optional

import yaml

from nefelibata.announcers.base import Announcement, Announcer, Scope, get_announcers
from nefelibata.constants import ANNOUNCEMENTS_FILENAME, PUBLISHINGS_FILENAME
from nefelibata.post import Post, get_posts
from nefelibata.publishers.base import Publisher, Publishing, get_publishers
from nefelibata.utils import get_config, load_yaml

_logger = logging.getLogger(__name__)


async def publish_site(
    name: str,
    publisher: Publisher,
    publishings: Dict[str, Publishing],
    since: Optional[datetime],
    force: bool = False,
) -> None:
    """
    Announce a site and store the result.
    """
    publishing = await publisher.publish(since, force)
    if publishing:
        publishings[name] = publishing


async def announce_site(
    name: str,
    announcer: Announcer,
    announcements: Dict[str, Announcement],
) -> None:
    """
    Announce a site and store the result.
    """
    announcement = await announcer.announce_site()
    if announcement:
        announcements[name] = announcement


async def announce_post(
    name: str,
    announcer: Announcer,
    post: Post,
    announcements: Dict[str, Announcement],
) -> None:
    """
    Announce a post and store the result.
    """
    announcement = await announcer.announce_post(post)
    if announcement:
        announcements[name] = announcement


async def save_announcements(
    post_directory: Path,
    announcements: Dict[str, Announcement],
) -> None:
    """
    Save modified post announcements.
    """
    path = post_directory / ANNOUNCEMENTS_FILENAME
    with open(path, "w", encoding="utf-8") as output:
        return yaml.dump(
            {name: announcement.dict() for name, announcement in announcements.items()},
            output,
        )


async def run(  # pylint: disable=too-many-locals
    root: Path,
    force: bool = False,
) -> None:
    """
    Publish blog.
    """
    _logger.info("Publishing blog")

    config = get_config(root)
    _logger.debug(config)

    # publish site
    publishings = load_yaml(root / PUBLISHINGS_FILENAME, Publishing)
    tasks = []
    for name, publisher in get_publishers(root, config).items():
        since = publishings[name].timestamp if name in publishings else None
        task = asyncio.create_task(
            publish_site(name, publisher, publishings, since, force),
        )
        tasks.append(task)

    await asyncio.gather(*tasks)

    # persist publishings
    with open(root / PUBLISHINGS_FILENAME, "w", encoding="utf-8") as output:
        yaml.dump(
            {name: publishing.dict() for name, publishing in publishings.items()},
            output,
        )

    # announcements
    tasks = []

    # announce site
    site_announcements = load_yaml(root / ANNOUNCEMENTS_FILENAME, Announcement)
    last_published = max(publishing.timestamp for publishing in publishings.values())
    announcers = get_announcers(root, config, Scope.SITE)
    for name, announcer in announcers.items():
        if (
            name in site_announcements
            and (
                site_announcements[name].timestamp
                + timedelta(seconds=site_announcements[name].grace_seconds)
            )
            >= last_published
        ):
            # already announced after last published
            _logger.info("Announcer %s is up-to-date", name)
            continue

        task = asyncio.create_task(announce_site(name, announcer, site_announcements))
        tasks.append(task)

    # announce posts
    modified_post_announcements: Dict[Path, Dict[str, Announcement]] = {}
    announcers = get_announcers(root, config, Scope.POST)
    for post in get_posts(root, config):
        path = post.path.parent / ANNOUNCEMENTS_FILENAME
        post_announcements = load_yaml(path, Announcement)
        post_announcers = {
            name: announcers[name]
            for name in post.announcers
            if name in announcers and name not in post_announcements
        }
        for name, announcer in post_announcers.items():
            task = asyncio.create_task(
                announce_post(name, announcer, post, post_announcements),
            )
            tasks.append(task)

        # store new announcements to persist later
        if post_announcers:
            modified_post_announcements[post.path] = post_announcements

    await asyncio.gather(*tasks)

    # persist new announcements
    tasks = []
    task = asyncio.create_task(save_announcements(root, site_announcements))
    for post_path, announcements in modified_post_announcements.items():
        task = asyncio.create_task(save_announcements(post_path.parent, announcements))
        tasks.append(task)

    await asyncio.gather(*tasks)



"""
Test ``nefelibata.cli.init``.
"""
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from nefelibata.cli import init
from nefelibata.constants import CONFIG_FILENAME


@pytest.mark.asyncio
async def test_run(
    mocker: MockerFixture,
    root: Path,
) -> None:
    """
    Test ``run``.
    """
    builder = mocker.MagicMock()
    builder.process_post = mocker.AsyncMock()
    builder.process_site = mocker.AsyncMock()

    mocker.patch("nefelibata.cli.init.get_config")
    mocker.patch("nefelibata.cli.init.get_builders", return_value={"builder": builder})

    _logger = mocker.patch("nefelibata.cli.init._logger")

    await init.run(root)

    builder.setup.assert_called()
    _logger.info.assert_called_with("Blog created!")

    assert (root / CONFIG_FILENAME).exists()
    assert (root / "posts/first/index.mkd").exists()

    # error if running again
    with pytest.raises(IOError) as excinfo:
        await init.run(root)
    assert str(excinfo.value) == "File /path/to/blog/nefelibata.yaml already exists!"

    # forcing an overwrite works, but fails for some reason with pyfakefs
    # await init.run(root, force=True)



"""
Tests for ``nefelibata.announcers.base``.
"""

from pathlib import Path
from typing import List, Type

import pytest
from pytest_mock import MockerFixture

from nefelibata.announcers.base import Announcer, Scope, get_announcers
from nefelibata.config import AnnouncerModel, BuilderModel, Config
from nefelibata.post import Post

from ..conftest import MockEntryPoint


def make_dummy_announcer(class_name: str, scopes: List[Scope]) -> Type[Announcer]:
    """
    Make a dummy ``Announcer`` derived class.
    """
    return type(class_name, (Announcer,), {"scopes": scopes})


@pytest.mark.asyncio
async def test_announcer(root: Path, config: Config, post: Post) -> None:
    """
    Test the base announcer methods.
    """
    announcer = Announcer(root, config, [])

    assert (await announcer.announce_post(post)) is None
    assert (await announcer.announce_site()) is None
    assert (await announcer.collect_post(post)) == {}
    assert (await announcer.collect_site()) == {}


def test_get_announcers(
    mocker: MockerFixture,
    make_entry_point: Type[MockEntryPoint],
    root: Path,
    config: Config,
) -> None:
    """
    Test ``get_announcers``.
    """
    DummyAnnouncer = make_dummy_announcer("DummyAnnouncer", [Scope.SITE])
    entry_points = [
        make_entry_point("dummy_announcer", DummyAnnouncer),
    ]
    mocker.patch(
        "nefelibata.announcers.base.iter_entry_points",
        return_value=entry_points,
    )
    mocker.patch(
        "nefelibata.announcers.base.get_builders",
    )

    config.builders = {
        "site_builder": BuilderModel(
            **{
                "plugin": "site_builder",
                "announce-on": ["dummy_announcer"],
                "publish-to": [],
                "home": "https://example.com/",
                "path": "generic",
            }
        ),
    }
    config.announcers = {
        "dummy_announcer": AnnouncerModel(
            **{
                "plugin": "dummy_announcer",
            }
        ),
    }
    announcers = get_announcers(root, config, Scope.SITE)
    assert len(announcers) == 1

    announcers = get_announcers(root, config, Scope.POST)
    assert len(announcers) == 0



"""
The configuration of a Nefelibata blog.
"""
# pylint: disable=too-few-public-methods

from typing import Dict, List

from pydantic import BaseModel, Extra, Field


class AuthorModel(BaseModel):
    """
    Model representing the author of the blog.
    """

    name: str
    url: str
    email: str
    note: str = ""


class SocialModel(BaseModel):
    """
    Model representing links to social media presence.
    """

    title: str
    url: str


class CategoryModel(BaseModel):
    """
    Model representing a category and the associated tags.
    """

    label: str
    description: str
    tags: List[str]


class BuilderModel(BaseModel):
    """
    Model representing a Nefelibata builder.
    """

    plugin: str
    announce_on: List[str] = Field(..., alias="announce-on")
    publish_to: List[str] = Field(..., alias="publish-to")
    home: str
    path: str

    class Config:
        """
        Allow extra attributes that are builder-specific.
        """

        extra = Extra.allow


class AssistantModel(BaseModel):
    """
    Model representing a Nefelibata assistant.
    """

    plugin: str

    class Config:
        """
        Allow extra attributes that are assistant-specific.
        """

        extra = Extra.allow


class AnnouncerModel(BaseModel):
    """
    Model representing a Nefelibata announcer.
    """

    plugin: str

    class Config:
        """
        Allow extra attributes that are announcer-specific.
        """

        extra = Extra.allow


class PublisherModel(BaseModel):
    """
    Model representing a Nefelibata publisher.
    """

    plugin: str

    class Config:
        """
        Allow extra attributes that are publisher-specific.
        """

        extra = Extra.allow


class Config(BaseModel):
    """
    Model representing the blog configuration.
    """

    title: str
    subtitle: str = ""
    author: AuthorModel
    language: str

    social: List[SocialModel]
    categories: Dict[str, CategoryModel]

    templates: Dict[str, List[str]]

    builders: Dict[str, BuilderModel]
    assistants: Dict[str, AssistantModel]
    announcers: Dict[str, AnnouncerModel]
    publishers: Dict[str, PublisherModel]



"""
A publisher that calls external commands.
"""
import logging
import os
import subprocess
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, List, Optional

from nefelibata.config import Config
from nefelibata.publishers.base import Publisher, Publishing

_logger = logging.getLogger(__name__)


class CommandPublisher(Publisher):

    """
    A publisher that calls external commands.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        root: Path,
        config: Config,
        path: str,
        post_commands: List[str],
        site_commands: List[str],
        **kwargs: Any,
    ):
        super().__init__(root, config, path, **kwargs)

        self.post_commands = post_commands
        self.site_commands = site_commands

    async def publish(
        self,
        since: Optional[datetime] = None,
        force: bool = False,
    ) -> Optional[Publishing]:
        build = self.root / "build" / self.path

        modified_files = list(self.find_modified_files(force, since))
        if not modified_files:
            return None

        os.chdir(build)
        for path in modified_files:
            for command in self.post_commands:
                _logger.info("Running command %s on file %s", command, path)
                subprocess.run(
                    command,
                    shell=True,
                    check=True,
                    env={"path": path.relative_to(build)},
                )

        for command in self.site_commands:
            _logger.info("Running command %s", command)
            subprocess.run(command, shell=True, check=True)

        return Publishing(timestamp=datetime.now(timezone.utc))



"""
Tests for ``nefelibata.announcers.mastodon``.
"""
# pylint: disable=invalid-name, redefined-outer-name

from datetime import datetime, timezone
from pathlib import Path

import pytest
import yaml
from freezegun import freeze_time
from pytest_mock import MockerFixture
from yarl import URL

from nefelibata.announcers.base import Announcement, Author, Interaction
from nefelibata.announcers.webmention import (
    Webmention,
    WebmentionAnnouncer,
    get_webmention_endpoint,
    send_webmention,
    update_webmention,
)
from nefelibata.builders.base import Builder
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_get_webmention_endpoint(mocker: MockerFixture) -> None:
    """
    Test ``get_webmention_endpoint``.
    """
    session = mocker.MagicMock()
    head_response = session.head.return_value.__aenter__.return_value
    head_response.headers = {
        "content-type": "text/html",
    }
    head_response.links = {
        "alternate": {
            "url": "/atom.xml",
        },
        "webmention": {
            "url": "/webmention.php",
        },
    }
    get_response = session.get.return_value.__aenter__.return_value
    get_response.text = mocker.AsyncMock()
    get_response.text.return_value = """<!doctype html>
<html lang="en">
  <head>
    <link rel="webmention" href="/webmention.php" />
  </head>
  <body>
  </body>
</html>"""

    # endpoint in Link header
    endpoint = await get_webmention_endpoint(
        session,
        URL("https://example.com/post/hello.php"),
    )
    assert endpoint == URL("https://example.com/webmention.php")

    # endpoint in HTML
    head_response.links = {}
    endpoint = await get_webmention_endpoint(
        session,
        URL("https://example.com/post/hello.php"),
    )
    assert endpoint == URL("https://example.com/webmention.php")

    # no endpoint in HTML
    get_response.text.return_value = """<!doctype html>
<html lang="en">
  <head>
  </head>
  <body>
  </body>
</html>"""
    endpoint = await get_webmention_endpoint(
        session,
        URL("https://example.com/post/hello.php"),
    )
    assert endpoint is None

    # unable to extract endpoint
    head_response.headers = {
        "content-type": "application/pdf",
    }
    endpoint = await get_webmention_endpoint(
        session,
        URL("https://example.com/post/hello.php"),
    )
    assert endpoint is None

    # gemini target
    head_response.links = {}
    endpoint = await get_webmention_endpoint(
        session,
        URL("gemini://example.com/post/hello.gmi"),
    )
    assert endpoint is None


@pytest.mark.asyncio
async def test_get_webmention_endpoint_unicode_error(mocker: MockerFixture) -> None:
    """
    Test ``get_webmention_endpoint`` when response is not text.
    """
    session = mocker.MagicMock()
    head_response = session.head.return_value.__aenter__.return_value
    head_response.headers = {
        "content-type": "text/html",
    }
    head_response.links = {}
    mocker.patch("nefelibata.announcers.webmention.UnicodeDecodeError", Exception)
    get_response = session.get.return_value.__aenter__.return_value
    get_response.text = mocker.AsyncMock()
    get_response.text.side_effect = Exception()

    endpoint = await get_webmention_endpoint(
        session,
        URL("https://example.com/post/hello.php"),
    )
    assert endpoint is None


@pytest.mark.asyncio
async def test_send_webmention(mocker: MockerFixture) -> None:
    """
    Test ``send_webmention``.
    """
    session = mocker.MagicMock()
    post_response = session.post.return_value.__aenter__.return_value
    post_response.raise_for_status = mocker.MagicMock()
    post_response.raise_for_status.return_value = None

    source = URL("https://alice.example.com/posts/one")
    target = URL("https://bob.example.com/posts/two")

    # no endpoint
    mocker.patch(
        "nefelibata.announcers.webmention.get_webmention_endpoint",
        return_value=None,
    )
    webmention = await send_webmention(session, source, target)
    assert webmention.dict() == {
        "source": "https://alice.example.com/posts/one",
        "target": "https://bob.example.com/posts/two",
        "status": "invalid",
        "location": None,
    }

    # 200 OK
    mocker.patch(
        "nefelibata.announcers.webmention.get_webmention_endpoint",
        return_value=URL("//bob.example.com/webmention.php"),
    )
    post_response.status = 200
    webmention = await send_webmention(session, source, target)
    assert webmention.dict() == {
        "source": "https://alice.example.com/posts/one",
        "target": "https://bob.example.com/posts/two",
        "status": "success",
        "location": None,
    }

    # 201 CREATED
    post_response.status = 201
    post_response.headers = {"location": "https://bob.example.com/webmention.php?id=42"}
    webmention = await send_webmention(session, source, target)
    assert webmention.dict() == {
        "source": "https://alice.example.com/posts/one",
        "target": "https://bob.example.com/posts/two",
        "status": "queue",
        "location": "https://bob.example.com/webmention.php?id=42",
    }

    # 202 ACCEPTED
    post_response.status = 202
    webmention = await send_webmention(session, source, target)
    assert webmention.dict() == {
        "source": "https://alice.example.com/posts/one",
        "target": "https://bob.example.com/posts/two",
        "status": "success",
        "location": None,
    }

    # error
    post_response.raise_for_status.side_effect = Exception("Error")
    mocker.patch("nefelibata.announcers.webmention.ClientResponseError", Exception)
    webmention = await send_webmention(session, source, target)
    assert webmention.dict() == {
        "source": "https://alice.example.com/posts/one",
        "target": "https://bob.example.com/posts/two",
        "status": "error",
        "location": None,
    }


@pytest.mark.asyncio
async def test_update_webmention(mocker: MockerFixture) -> None:
    """
    Test ``update_webmention``.
    """
    session = mocker.MagicMock()
    get_response = session.get.return_value.__aenter__.return_value

    source = URL("https://alice.example.com/posts/one")
    target = URL("https://bob.example.com/posts/two")
    location = URL("https://bob.example.com/webmention.php?id=42")

    get_response.ok = True
    webmention = await update_webmention(session, source, target, location)
    assert webmention.dict() == {
        "source": "https://alice.example.com/posts/one",
        "target": "https://bob.example.com/posts/two",
        "status": "success",
        "location": None,
    }

    get_response.ok = False
    webmention = await update_webmention(session, source, target, location)
    assert webmention.dict() == {
        "source": "https://alice.example.com/posts/one",
        "target": "https://bob.example.com/posts/two",
        "status": "queue",
        "location": None,
    }


@pytest.mark.asyncio
async def test_announcer_announce(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test announcing posts.
    """
    gemini_builder = Builder(root, config, "gemini://example.com/", "gemini")
    gemini_builder.extension = ".gmi"
    html_builder = Builder(root, config, "https://example.com/", "www")
    html_builder.extension = ".html"

    send_webmention = mocker.AsyncMock(
        side_effect=[
            Webmention(
                source="gemini://example.com/first/index.html",
                target="https://nefelibata.readthedocs.io/",
                status="invalid",
            ),
            Webmention(
                source="https://example.com/first/index.html",
                target="https://nefelibata.readthedocs.io/",
                status="queue",
                location="https://bob.example.com/webmention.php?id=42",
            ),
        ],
    )
    mocker.patch("nefelibata.announcers.webmention.send_webmention", send_webmention)

    announcer = WebmentionAnnouncer(
        root,
        config,
        builders=[gemini_builder, html_builder],
    )

    announcement = await announcer.announce_post(post)
    assert announcement is None

    path = post.path.parent / "webmentions.yaml"
    with open(path, encoding="utf-8") as input_:
        webmentions = yaml.load(input_, Loader=yaml.SafeLoader)
    assert webmentions == {
        "gemini://example.com/first/index.gmi => https://nefelibata.readthedocs.io/": {
            "location": None,
            "source": "gemini://example.com/first/index.html",
            "status": "invalid",
            "target": "https://nefelibata.readthedocs.io/",
        },
        "https://example.com/first/index.html => https://nefelibata.readthedocs.io/": {
            "location": "https://bob.example.com/webmention.php?id=42",
            "source": "https://example.com/first/index.html",
            "status": "queue",
            "target": "https://nefelibata.readthedocs.io/",
        },
    }

    update_webmention = mocker.AsyncMock(
        return_value=Webmention(
            source="https://example.com/first/index.html",
            target="https://nefelibata.readthedocs.io/",
            status="success",
            location=None,
        ),
    )
    mocker.patch(
        "nefelibata.announcers.webmention.update_webmention",
        update_webmention,
    )

    with freeze_time("2021-01-01T00:00:00Z"):
        announcement = await announcer.announce_post(post)
    assert announcement == Announcement(
        url="first/index",
        timestamp=datetime(2021, 1, 1, tzinfo=timezone.utc),
        grace_seconds=0,
    )

    path = post.path.parent / "webmentions.yaml"
    with open(path, encoding="utf-8") as input_:
        webmentions = yaml.load(input_, Loader=yaml.SafeLoader)
    assert webmentions == {
        "gemini://example.com/first/index.gmi => https://nefelibata.readthedocs.io/": {
            "location": None,
            "source": "gemini://example.com/first/index.html",
            "status": "invalid",
            "target": "https://nefelibata.readthedocs.io/",
        },
        "https://example.com/first/index.html => https://nefelibata.readthedocs.io/": {
            "location": None,
            "source": "https://example.com/first/index.html",
            "status": "success",
            "target": "https://nefelibata.readthedocs.io/",
        },
    }


@pytest.mark.asyncio
async def test_announcer_collect(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test collecting posts.
    """
    gemini_builder = Builder(root, config, "gemini://example.com/", "gemini")
    gemini_builder.extension = ".gmi"
    html_builder = Builder(root, config, "https://example.com/", "www")
    html_builder.extension = ".html"

    mocker.patch("nefelibata.announcers.webmention.ClientResponseError", Exception)
    get = mocker.patch("nefelibata.announcers.webmention.ClientSession.get")
    get_response = get.return_value.__aenter__.return_value
    get_response.raise_for_status = mocker.MagicMock()
    get_response.raise_for_status.side_effect = [
        Exception("Gemini not supported"),
        None,
        Exception("Gemini not supported"),
        None,
    ]

    announcer = WebmentionAnnouncer(
        root,
        config,
        builders=[gemini_builder, html_builder],
    )

    get_response.json = mocker.AsyncMock(return_value={"children": []})
    interactions = await announcer.collect_post(post)
    assert interactions == {}

    entries = [
        {
            "type": "entry",
            "wm-id": 1,
            "wm-source": "https://alice.example.com/posts/one",
            "name": "This is the title",
            "summary": {"value": "This is the summary"},
            "content": {"text": "This is the post content."},
            "published": "2021-01-01T00:00:00Z",
            "author": {
                "name": "Alice Doe",
                "url": "https://alice.example.com/",
                "photo": "https://alice.example.com/photo.jpg",
                "note": "My name is Alice",
            },
            "wm-property": "in-reply-to",
        },
        {
            "type": "entry",
            "wm-id": 2,
            "wm-source": "https://bob.example.com/posts/two",
            "summary": {"value": "This is the summary"},
            "content": {"text": "This is the post content."},
            "published": "2021-01-01T00:00:00Z",
            "author": {
                "name": "Bob Doe",
                "url": "https://bob.example.com/",
                "photo": "https://bob.example.com/photo.jpg",
            },
            "wm-property": "like-of",
        },
        {"type": "invalid"},
    ]

    get_response.json = mocker.AsyncMock(return_value={"children": entries})
    interactions = await announcer.collect_post(post)
    assert interactions == {
        1: Interaction(
            id="1",
            name="This is the title",
            summary="This is the summary",
            content="This is the post content.",
            published=datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            updated=None,
            author=Author(
                name="Alice Doe",
                url="https://alice.example.com/",
                avatar="https://alice.example.com/photo.jpg",
                note="My name is Alice",
            ),
            url="https://alice.example.com/posts/one",
            in_reply_to=None,
            type="reply",
        ),
        2: Interaction(
            id="2",
            name="https://bob.example.com/posts/two",
            summary="This is the summary",
            content="This is the post content.",
            published=datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            updated=None,
            author=Author(
                name="Bob Doe",
                url="https://bob.example.com/",
                avatar="https://bob.example.com/photo.jpg",
                note="",
            ),
            url="https://bob.example.com/posts/two",
            in_reply_to=None,
            type="like",
        ),
    }



"""
An announcer for webmentions.
"""

import asyncio
import logging
from datetime import datetime, timezone
from typing import Dict, Literal, Optional, TypedDict

import dateutil.parser
from aiohttp import ClientResponseError, ClientSession
from bs4 import BeautifulSoup
from pydantic import BaseModel
from yarl import URL

from nefelibata.announcers.base import (
    Announcement,
    Announcer,
    Author,
    Interaction,
    Scope,
)
from nefelibata.post import Post, extract_links
from nefelibata.utils import update_yaml

_logger = logging.getLogger(__name__)


# map from wm-property to ``Interaction.type``
INTERACTION_TYPES = {
    "in-reply-to": "reply",
    "mention-of": "mention",
    "like-of": "like",
}


class Webmention(BaseModel):
    """
    A webmention.
    """

    source: str
    target: str
    status: Literal["success", "queue", "invalid", "error"]
    location: Optional[str] = None


class WebmentionType(TypedDict, total=False):
    """
    Type for ``Webmention.dict()``.
    """

    source: str
    target: str
    status: Literal["success", "queue", "invalid", "error"]
    location: str


async def get_webmention_endpoint(session: ClientSession, target: URL) -> Optional[URL]:
    """
    Given a target URL, find the webmention endpoint, if any.
    """
    if target.scheme not in {"http", "https"}:
        return None

    async with session.head(target) as response:
        for rel, params in response.links.items():
            if "webmention" in rel.split(" "):
                return target.join(URL(params["url"]))

        if "text/html" not in response.headers.get("content-type", ""):
            return None

    async with session.get(target) as response:
        try:
            html = await response.text()
        except UnicodeDecodeError:
            return None

    soup = BeautifulSoup(html, "html.parser")
    link = soup.find(rel="webmention", href=True)
    if link:
        return target.join(URL(link["href"]))

    return None


async def send_webmention(
    session: ClientSession,
    source: URL,
    target: URL,
) -> Webmention:
    """
    Send a webmention to a target.
    """
    endpoint = await get_webmention_endpoint(session, target)
    if not endpoint:
        _logger.info("No endpoint found")
        return Webmention(
            source=str(source),
            target=str(target),
            status="invalid",
        )

    _logger.info("Sending webmention to %s", endpoint)
    payload = {"source": source, "target": target}
    async with session.post(endpoint, data=payload) as response:
        try:
            response.raise_for_status()
        except ClientResponseError:
            _logger.error("Error sending webmention")
            return Webmention(
                source=str(source),
                target=str(target),
                status="error",
            )

        if response.status == 201:
            return Webmention(
                source=str(source),
                target=str(target),
                status="queue",
                location=response.headers["location"],
            )

    return Webmention(
        source=str(source),
        target=str(target),
        status="success",
    )


async def update_webmention(
    session: ClientSession,
    source: URL,
    target: URL,
    location: URL,
) -> Webmention:
    """
    Update the status on a queued webmention.
    """
    async with session.get(location) as response:
        status = "success" if response.ok else "queue"

    return Webmention(
        source=str(source),
        target=str(target),
        status=status,
    )


async def collect_webmentions(
    session: ClientSession,
    source: URL,
    target: URL,
    webmentions: Dict[str, WebmentionType],
) -> None:
    """
    Helper function to collect webmentions.
    """
    key = f"{source} => {target}"
    if key not in webmentions:
        webmention = await send_webmention(session, source, target)
        webmentions[key] = webmention.dict()
    elif webmentions[key]["status"] == "queue":
        location = webmentions[key]["location"]
        webmention = await update_webmention(session, source, target, location)
        webmentions[key] = webmention.dict()


class WebmentionAnnouncer(Announcer):

    """
    An announcer for Webmention (https://indieweb.org/Webmention).
    """

    scopes = [Scope.POST]

    async def announce_post(self, post: Post) -> Optional[Announcement]:
        path = post.path.parent / "webmentions.yaml"

        tasks = []
        with update_yaml(path) as webmentions:
            async with ClientSession() as session:
                for target in extract_links(post):
                    for builder in self.builders:
                        source = builder.absolute_url(post)
                        task = asyncio.create_task(
                            collect_webmentions(session, source, target, webmentions),
                        )
                        tasks.append(task)

                await asyncio.gather(*tasks)

        if any(webmention["status"] == "queue" for webmention in webmentions.values()):
            # return nothing so we can check queued webmentions on next publish
            return None

        return Announcement(
            url=post.url,
            timestamp=datetime.now(timezone.utc),
        )

    async def collect_post(self, post: Post) -> Dict[str, Interaction]:
        interactions: Dict[str, Interaction] = {}

        async with ClientSession() as session:
            for builder in self.builders:
                target = builder.absolute_url(post)
                payload = {"target": target}

                async with session.get(
                    "https://webmention.io/api/mentions.jf2",
                    data=payload,
                ) as response:
                    try:
                        response.raise_for_status()
                    except ClientResponseError:
                        _logger.error("Error fetching webmentions")
                        continue

                    payload = await response.json()

                for entry in payload["children"]:
                    if entry["type"] != "entry":
                        continue

                    interactions[entry["wm-id"]] = Interaction(
                        id=entry["wm-id"],
                        name=entry.get("name", entry["wm-source"]),
                        summary=entry.get("summary", {}).get("value"),
                        content=entry["content"]["text"],
                        published=dateutil.parser.parse(entry["published"]),
                        updated=None,
                        author=Author(
                            name=entry["author"]["name"],
                            url=entry["author"]["url"],
                            avatar=entry["author"]["photo"],
                            note=entry["author"].get("note", ""),
                        ),
                        url=entry["wm-source"],
                        in_reply_to_id=None,
                        type=INTERACTION_TYPES[entry["wm-property"]],
                    )

        return interactions



"""
Create a new blog skeleton.
"""
import logging
import os
import shutil
from pathlib import Path

from pkg_resources import resource_filename, resource_listdir

from nefelibata.builders.base import get_builders
from nefelibata.utils import get_config

_logger = logging.getLogger(__name__)


async def run(root: Path, force: bool = False) -> None:
    """
    Create a new blog skeleton.
    """
    resources = sorted(resource_listdir("nefelibata", "templates/skeleton"))
    for resource in resources:
        origin = Path(
            resource_filename(
                "nefelibata",
                os.path.join("templates/skeleton", resource),
            ),
        )
        target = root / resource
        if target.exists() and not force:
            resource_type = "Directory" if origin.is_dir() else "File"
            raise IOError(f"{resource_type} {target} already exists!")
        if origin.is_dir():
            target.mkdir(parents=True, exist_ok=True)
            shutil.copytree(origin, target, dirs_exist_ok=True)
        else:
            target.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(origin, target)

    # create templates
    config = get_config(root)
    for builder in get_builders(root, config).values():
        builder.setup()

    _logger.info("Blog created!")



"""
Tests for ``nefelibata.builders.gemini``.
"""
# pylint: disable=invalid-name
from pathlib import Path
from typing import Any, Dict

import pytest
from freezegun import freeze_time
from marko import Markdown
from pytest_mock import MockerFixture

from nefelibata.builders.gemini import GeminiBuilder, GeminiRenderer
from nefelibata.config import Config, SocialModel
from nefelibata.enclosure import Enclosure
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_builder_setup(root: Path, config: Config) -> None:
    """
    Test that templates and build directory are created.
    """
    template_directory = root / "templates/builders/gemini"
    build_directory = root / "build/gemini"
    assert not template_directory.exists()
    assert not build_directory.exists()

    GeminiBuilder(root, config, "gemini://localhost:1965")
    last_update: Dict[Any, Any] = {}
    assert template_directory.exists()
    for file in ("index.gmi", "feed.gmi", "post.gmi"):
        assert (template_directory / file).exists()
        last_update[file] = (template_directory / file).stat().st_mtime
    assert build_directory.exists()
    last_update[build_directory] = build_directory.stat().st_mtime

    # on the second call the files and directories should be unmodified
    GeminiBuilder(root, config, "gemini://localhost:1965")
    for file in ("index.gmi", "feed.gmi", "post.gmi"):
        assert (template_directory / file).stat().st_mtime == last_update[file]
    assert build_directory.stat().st_mtime == last_update[build_directory]


@pytest.mark.asyncio
async def test_builder_post(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test ``process_post``.
    """
    _logger = mocker.patch("nefelibata.builders.base._logger")

    enclosure_path = post.path.parent / "picture.jpg"
    enclosure_path.touch()
    post.enclosures.append(
        Enclosure(
            path=enclosure_path,
            description="A photo",
            type="image/jpeg",
            length=666,
            href="first/picture.jpg",
        ),
    )

    builder = GeminiBuilder(root, config, "gemini://localhost:1965")
    with freeze_time("2021-01-02T00:00:00Z"):
        await builder.process_post(post)

    post_path = root / "build/gemini/first/index.gmi"

    # test that file was created
    assert post_path.exists()
    _logger.info.assert_has_calls(
        [
            mocker.call("Creating %s post", "Gemini"),
            mocker.call("Copying enclosure %s", enclosure_path),
        ],
    )
    with open(post_path, encoding="utf-8") as input_:
        content = input_.read()
    assert (
        content
        == """# This is your first post

# Welcome

This is your first post. It should be written using Markdown.

Read more about Nefelibata.

=> https://nefelibata.readthedocs.io/ Nefelibata

# Enclosures

=> gemini://localhost:1965/first/picture.jpg A photo

# Tags

=> gemini://localhost:1965/tags/blog.gmi blog
=> gemini://localhost:1965/tags/welcome.gmi welcome

# Categories

=> gemini://localhost:1965/categories/stem.gmi STEM

# About

Published on 2020-12-31 16:00:00-08:00 by Beto Dealmeida <roberto@dealmeida.net>.

=> gemini://localhost:1965 Go back to the main index"""
    )

    # call again, test that file is up-to-date
    _logger.reset_mock()
    last_update = post_path.stat().st_mtime
    with freeze_time("2021-01-03T00:00:00Z"):
        await builder.process_post(post)
    assert post_path.stat().st_mtime == last_update
    _logger.debug.assert_called_with("Post %s is up-to-date, nothing to do", post_path)

    # call again, forcing a rebuild
    _logger.reset_mock()
    last_update = post_path.stat().st_mtime
    with freeze_time("2021-01-04T00:00:00Z"):
        await builder.process_post(post, force=True)
    assert post_path.stat().st_mtime > last_update
    _logger.info.assert_has_calls(
        [
            mocker.call("Creating %s post", "Gemini"),
            mocker.call("Copying enclosure %s", enclosure_path),
        ],
    )


@pytest.mark.asyncio
async def test_builder_site(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test ``process_site``.
    """
    _logger = mocker.patch("nefelibata.builders.base._logger")
    mocker.patch("nefelibata.builders.base.get_posts", return_value=[post])

    config.social = [SocialModel(title="Mastodon", url="https://2c.taoetc.org/@beto")]
    builder = GeminiBuilder(root, config, "gemini://localhost:1965")
    with freeze_time("2021-01-02T00:00:00Z"):
        await builder.process_site()

    assets_directory = root / "build/gemini"
    assets = (
        "index.gmi",
        "feed.gmi",
        "tags/welcome.gmi",
        "tags/blog.gmi",
        "categories/stem.gmi",
    )

    # test that files were created
    last_update: Dict[Any, Any] = {}
    for asset in assets:
        assert (assets_directory / asset).exists()
        last_update[asset] = (assets_directory / asset).stat().st_mtime
    _logger.info.assert_has_calls(
        [
            mocker.call("Creating %s", Path("/path/to/blog/build/gemini/index.gmi")),
            mocker.call("Creating %s", Path("/path/to/blog/build/gemini/feed.gmi")),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/gemini/tags/blog.gmi"),
            ),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/gemini/tags/welcome.gmi"),
            ),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/gemini/categories/stem.gmi"),
            ),
        ],
    )
    with open(assets_directory / "index.gmi", encoding="utf-8") as input_:
        content = input_.read()
    space = " "
    assert (
        content
        == f"""# 道&c.: Musings about the path and other things

This is the Gemini capsule of Beto Dealmeida.

=> https://taoetc.org/ Website
=> mailto://roberto@dealmeida.net Email address
=> gemini://localhost:1965/feed.gmi Gemlog

## Posts

=> gemini://localhost:1965/first/index.gmi 2020-12-31 16:00:00-08:00 — This is your first post{space}

## Links

=> https://2c.taoetc.org/@beto Mastodon

Crafted with ❤️ using Nefelibata

=> https://nefelibata.readthedocs.io/ Nefelibata"""
    )
    with open(assets_directory / "feed.gmi", encoding="utf-8") as input_:
        content = input_.read()
    assert (
        content
        == """# 道&c.

=> gemini://localhost:1965/first/index.gmi 2020-12-31 — This is your first post
"""
    )

    # call again, test that file is up-to-date
    _logger.reset_mock()
    with freeze_time("2021-01-03T00:00:00Z"):
        await builder.process_site()
    for asset in assets:
        assert (assets_directory / asset).stat().st_mtime == last_update[asset]
    _logger.debug.assert_has_calls(
        [
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/gemini/index.gmi"),
            ),
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/gemini/feed.gmi"),
            ),
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/gemini/tags/blog.gmi"),
            ),
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/gemini/tags/welcome.gmi"),
            ),
            mocker.call(
                "File %s is up-to-date, nothing to do",
                Path("/path/to/blog/build/gemini/categories/stem.gmi"),
            ),
        ],
    )

    # call again, forcing a rebuild
    _logger.reset_mock()
    with freeze_time("2021-01-04T00:00:00Z"):
        await builder.process_site(force=True)
    for asset in assets:
        assert (assets_directory / asset).stat().st_mtime > last_update[asset]
    _logger.info.assert_has_calls(
        [
            mocker.call("Creating %s", Path("/path/to/blog/build/gemini/index.gmi")),
            mocker.call("Creating %s", Path("/path/to/blog/build/gemini/feed.gmi")),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/gemini/tags/blog.gmi"),
            ),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/gemini/tags/welcome.gmi"),
            ),
            mocker.call(
                "Creating %s",
                Path("/path/to/blog/build/gemini/categories/stem.gmi"),
            ),
        ],
    )


def test_gemini_renderer() -> None:
    """
    Test ``GeminiRenderer``.
    """
    double_space = "  "

    gemini = Markdown(renderer=GeminiRenderer)

    assert (
        gemini.convert(
            f"""
# Header 1

## Header 2

### Header 3

#### Header 4

- Mercury
- Gemini
  - Apollo

> This is{double_space}
> A multiline{double_space}
> blockquote{double_space}

*This* is a **paragraph**. With `code`.

This is an [inline link](https://example.com/). This is [another](https://example.org/).

```
This is some code.
```

End.
""",
        )
        == f"""
# Header 1

## Header 2

### Header 3

### Header 4

* Mercury
* Gemini
* Apollo

> This is
> A multiline
> blockquote{double_space}

This is a paragraph. With code.

This is an inline link. This is another.

=> https://example.com/ inline link
=> https://example.org/ another

```
This is some code.
```

End.
"""
    )


def test_gemini_renderer_link_ref_def() -> None:
    """
    Test rendering a link definition reference.
    """
    gemini = Markdown(renderer=GeminiRenderer)

    assert (
        gemini.convert(
            """
Hi, here's my [thing that I just casually mention][tt] sometimes.

[tt]: gemini://my.boring/url "I like this link"
            """,
        )
        == """
Hi, here's my thing that I just casually mention sometimes.

=> gemini://my.boring/url I like this link


"""
    )


def test_gemini_renderer_links() -> None:
    """
    Test rendering links.
    """
    gemini = Markdown(renderer=GeminiRenderer)
    assert (
        gemini.convert(
            """
This is a paragraph with [a link](https://example.com/). It is a good paragraph.

This one [also has a link](https://example.net/). It also has some *emphasis* and **bold**.

And this one is a special paragraph because it has not just [one](https://example.com/one),
but [two](https://example.com/two) links. That is quite a lot.

This other paragraph also has [a link][ref], using a reference. It's very fancy.

[ref]: https://example.org/foo "This is foo"

This one has an auto link <https://example.com/autolink>.

That is it.
            """,
        )
        == """
This is a paragraph with a link. It is a good paragraph.

=> https://example.com/ a link

This one also has a link. It also has some emphasis and bold.

=> https://example.net/ also has a link

And this one is a special paragraph because it has not just one,
but two links. That is quite a lot.

=> https://example.com/one one
=> https://example.com/two two

This other paragraph also has a link, using a reference. It's very fancy.

=> https://example.org/foo This is foo


This one has an auto link https://example.com/autolink.

=> https://example.com/autolink https://example.com/autolink

That is it.

"""
    )


def test_gemini_renderer_padding_after_link_ref_def() -> None:
    """
    Test the padding after a link reference definition.

    Ideally we shouldn't be left with an empty line.
    """
    gemini = Markdown(renderer=GeminiRenderer)
    assert (
        gemini.convert(
            """
This other paragraph also has [a link][ref], using a reference. It's very fancy.

[ref]: https://example.org/foo "This is foo"

That is it.
""",
        )
        == """
This other paragraph also has a link, using a reference. It's very fancy.

=> https://example.org/foo This is foo


That is it.
"""
    )


def test_gemini_renderer_image() -> None:
    """
    Test rendering images.
    """
    gemini = Markdown(renderer=GeminiRenderer)
    assert (
        gemini.convert(
            """
This is a paragraph.

![Alt text](https://assets.digitalocean.com/articles/alligator/boo.svg "a title")

![Alt text](https://assets.digitalocean.com/articles/alligator/boo.svg)

That is it.
""",
        )
        == """
This is a paragraph.

=> https://assets.digitalocean.com/articles/alligator/boo.svg a title

=> https://assets.digitalocean.com/articles/alligator/boo.svg Alt text

That is it.
"""
    )



"""
An FTP publisher.
"""
import logging
from datetime import datetime, timezone
from ftplib import FTP, FTP_TLS, error_perm
from pathlib import Path
from typing import Any, Optional

from nefelibata.config import Config
from nefelibata.publishers.base import Publisher, Publishing

_logger = logging.getLogger(__name__)


class FTPPublisher(Publisher):

    """
    A publisher that uploads the blog to an FTP server.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        root: Path,
        config: Config,
        path: str,
        hostname: str,
        username: str = "",
        password: str = "",
        basedir: str = "",
        use_tls: bool = False,
        **kwargs: Any,
    ):
        super().__init__(root, config, path, **kwargs)

        self.hostname = hostname
        self.username = username
        self.password = password
        self.basedir = basedir
        self.use_tls = use_tls

    async def publish(
        self,
        since: Optional[datetime] = None,
        force: bool = False,
    ) -> Optional[Publishing]:
        build = self.root / "build" / self.path

        FTPClass = FTP_TLS if self.use_tls else FTP
        with FTPClass(self.hostname, self.username, self.password) as ftp:
            if self.use_tls:
                ftp.prot_p()  # pylint: disable=no-member

            if self.basedir:
                ftp.cwd(self.basedir)
            pwd = basedir = Path(ftp.pwd())

            # get files that need to be published, sorted to minimize ``CWD`` calls
            modified_files = sorted(self.find_modified_files(force, since))

            for path in modified_files:
                relative_directory = path.relative_to(build).parent
                if relative_directory != pwd.relative_to(basedir):
                    if pwd != basedir:
                        ftp.cwd(str(basedir))
                    for directory in relative_directory.parts:
                        try:
                            ftp.cwd(directory)
                        except error_perm:
                            ftp.mkd(directory)
                            ftp.cwd(directory)
                    pwd = basedir / relative_directory

                _logger.info("Uploading %s", path)
                with open(path, "rb") as input_:
                    ftp.storbinary(f"STOR {path.name}", input_)

        if modified_files:
            return Publishing(timestamp=datetime.now(timezone.utc))

        return None



"""
Test ``nefelibata.cli.publish``.
"""
# pylint: disable=unused-argument

from datetime import datetime
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from nefelibata.announcers.base import Announcement
from nefelibata.cli import publish
from nefelibata.config import Config
from nefelibata.post import Post
from nefelibata.publishers.base import Publishing


@pytest.mark.asyncio
async def test_run(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test ``publish``.
    """
    publisher = mocker.MagicMock()
    publisher.publish = mocker.AsyncMock(
        side_effect=[
            Publishing(timestamp=datetime(2021, 1, 1)),
            None,
            Publishing(timestamp=datetime(2021, 1, 3)),
        ],
    )

    announcer = mocker.MagicMock()
    announcer.announce_post = mocker.AsyncMock(
        side_effect=[
            None,
            Announcement(
                url="https://host1.example.com/",
                timestamp=datetime(2021, 1, 1),
            ),
        ],
    )
    announcer.announce_site = mocker.AsyncMock(
        side_effect=[
            Announcement(
                url="https://host2.example.com/",
                timestamp=datetime(2021, 1, 2),
            ),
            None,
        ],
    )

    mocker.patch(
        "nefelibata.cli.publish.get_announcers",
        return_value={"announcer": announcer},
    )
    mocker.patch(
        "nefelibata.cli.publish.get_publishers",
        return_value={"publisher": publisher},
    )

    # On the first publish we should announce site and post.
    await publish.run(root)
    publisher.publish.assert_called_with(None, False)
    announcer.announce_site.assert_called_with()
    announcer.announce_post.assert_called()

    # Publish again, should have a ``since`` value. Because ``publish``
    # returns ``None`` the second time we shouldn't announce the site.
    # And because the first time ``publish_post`` returned ``None``,
    # it should try again now and be called.
    announcer.announce_site.reset_mock()
    announcer.announce_post.reset_mock()
    await publish.run(root)
    publisher.publish.assert_called_with(datetime(2021, 1, 1), False)
    announcer.announce_site.assert_not_called()
    announcer.announce_post.assert_called()

    # Publish again. This time ``publish`` returns a new data, so we
    # expect to call ``publish_site``. Because the post was already
    # published last time it shouldn't be announced this time.
    announcer.announce_post.reset_mock()
    await publish.run(root)
    publisher.publish.assert_called_with(datetime(2021, 1, 1), False)
    announcer.announce_site.assert_called_with()
    announcer.announce_post.assert_not_called()



"""
An announcer for gemlogs.
"""
import logging
import re
import ssl
import urllib.parse
from collections import defaultdict
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional

from aiogemini.client import Client
from aiogemini.security import TOFUContext
from yarl import URL

from nefelibata.announcers.base import Announcement, Announcer, Interaction, Scope
from nefelibata.builders.base import Builder
from nefelibata.config import Config
from nefelibata.post import get_posts

_logger = logging.getLogger(__name__)


class GemlogAnnouncer(Announcer):

    """
    A generic gemlog announcer.

    This requires the Gemini builder, and assumes that the Gemlog feed is
    at ``{builder.home}/feed{builder.extension}``.
    """

    scopes = [Scope.SITE]

    # Replace these with values from the place where you want to announce the
    # feed.
    name = ""
    url = "gemini://example.com/"
    submit_url = "gemini://example.com/submit?"
    # Set up an expiration to prevent the announcement to made too frequently;
    # eg, for CAPCOM we only need announce the feed once, while for Antenna we
    # need to announce it on every change.
    grace_seconds = 0

    # Replace with a specific logger to have better logs
    logger = _logger

    def __init__(
        self, root: Path, config: Config, builders: List[Builder], **kwargs: Any
    ):
        super().__init__(root, config, builders, **kwargs)

        self.client = Client(TOFUContext({}))

    async def announce_site(self) -> Optional[Announcement]:
        """
        Send the link.
        """
        for builder in self.builders:
            if builder.home.scheme != "gemini":
                raise Exception(
                    f"{self.name} announcer only works with `gemini://` builds",
                )

            feed_url = urllib.parse.quote_plus(
                f"{builder.home}/feed{builder.extension}",
            )
            url = self.submit_url + feed_url

            self.logger.info("Announcing feed %s to %s", feed_url, self.name)
            await self.client.get(URL(url))

        return Announcement(
            url=self.url,
            timestamp=datetime.now(timezone.utc),
            grace_seconds=self.grace_seconds,
        )

    async def collect_site(self) -> Dict[Path, Dict[str, Interaction]]:
        """
        Return replies to posts.

        This is done by scraping the capsule and searching for "Re: " posts.
        """
        response = await self.client.get(URL(self.url))
        payload = await response.read()
        content = payload.decode("utf-8")

        posts = get_posts(self.root, self.config)

        interactions: Dict[Path, Dict[str, Interaction]] = defaultdict(dict)
        for line in content.split("\n"):
            if not line.startswith("=>"):
                continue

            _, url, name = re.split(r"\s+", line, 2)
            for post in posts:
                reply = f"Re: {post.title}"
                if reply in name and await self._link_in_post(post.url, url):
                    id_ = f"reply,{url}"
                    interactions[post.path][id_] = Interaction(
                        id=id_,
                        name=name,
                        url=url,
                        type="reply",
                    )

        return interactions

    async def _link_in_post(self, post_url: str, url: str) -> bool:
        """
        Check that a given URL actually links to the post URL.
        """
        try:
            response = await self.client.get(URL(url))
        except ssl.SSLCertVerificationError:
            return True

        payload = await response.read()
        content = payload.decode("utf-8")

        for line in content.split("\n"):
            if line.startswith("=>") and post_url in line:
                return True

        return False



"""
An announcer that saves the post and blog in https://archive.org/.
"""
import logging
from datetime import datetime, timezone
from typing import Optional

from nefelibata.announcers.base import Announcement, Announcer, Scope
from nefelibata.post import Post
from nefelibata.utils import archive_urls

_logger = logging.getLogger(__name__)


class ArchiveBlogAnnouncer(Announcer):

    """
    An announcer that saves the post and blog in https://archive.org/.
    """

    scopes = [Scope.POST, Scope.SITE]

    async def announce_post(self, post: Post) -> Optional[Announcement]:
        urls = [builder.absolute_url(post) for builder in self.builders]
        saved_urls = await archive_urls(urls)
        if saved_urls:
            return Announcement(
                url="https://web.archive.org/save/",
                timestamp=datetime.now(timezone.utc),
            )
        return None

    async def announce_site(self) -> Optional[Announcement]:
        urls = [builder.home for builder in self.builders]
        saved_urls = await archive_urls(urls)
        if saved_urls:
            return Announcement(
                url="https://web.archive.org/save/",
                timestamp=datetime.now(timezone.utc),
            )
        return None



"""
Base class for builders.
"""
import logging
import shutil
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List

from jinja2 import Environment, FileSystemLoader, select_autoescape
from pkg_resources import iter_entry_points, resource_filename, resource_listdir
from yarl import URL

from nefelibata import __version__
from nefelibata.config import Config
from nefelibata.post import Post, get_posts

_logger = logging.getLogger(__name__)


class Builder:

    """
    A post builder.
    """

    # The name of the builder; this should match the name of the entry point.
    name = ""

    # A display name for the builder, eg, "Gemini".
    label = ""

    # The default extension of the templates and the generated files, including
    # the period (if any).
    extension = ""

    # The directory where the templates live inside ``templates/builders/{name}/``.
    template_base = ""

    # A list of templates that should be processed when building the site.
    site_templates: List[str] = []

    def __init__(
        self, root: Path, config: Config, home: str, path: str = "", **kwargs: Any
    ):
        self.root = root
        self.config = config
        self.path = path or self.name
        self.home = URL(home)
        self.kwargs = kwargs

        self.env = self.get_environment()

    def setup(self) -> None:
        """
        Create templates and build directory.

        This method is used to create templates required by the builder, as
        well as the directory where the build will be stored. It's called
        when the blog is first initialized (``nb init``), but also every time
        the plugin is instantiated, in case the user configures a builder
        after the blog initialization.
        """
        # create templates
        template_directory = Path("templates/builders") / self.name
        resources = resource_listdir("nefelibata", str(template_directory))
        for resource in resources:
            origin = Path(
                resource_filename("nefelibata", str(template_directory / resource)),
            )
            target = self.root / template_directory / resource
            if target.exists():
                continue
            _logger.info("Creating %s", template_directory / resource)
            if origin.is_dir():
                target.mkdir(parents=True, exist_ok=True)
                shutil.copytree(origin, target, dirs_exist_ok=True)
            else:
                target.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(origin, target)

        # create build directories
        build_directory = self.root / "build" / self.path
        if not build_directory.exists():
            build_directory.mkdir(parents=True)

        tags_directory = self.root / "build" / self.path / "tags"
        if not tags_directory.exists():
            tags_directory.mkdir()

        categories_directory = self.root / "build" / self.path / "categories"
        if not categories_directory.exists():
            categories_directory.mkdir()

    def absolute_url(self, post: Post) -> URL:
        """
        Return the absolute URL for a post.
        """
        return self.home / f"{post.url}{self.extension}"

    @staticmethod
    def render(content: str) -> str:
        """
        Render the post content.
        """
        return content

    def get_environment(self) -> Environment:
        """
        Load a Jinja2 environment pointing to the templats.
        """
        return Environment(
            loader=FileSystemLoader(str(self.root / "templates/builders" / self.name)),
            lstrip_blocks=True,
            trim_blocks=True,
            autoescape=select_autoescape(["html", "xml"]),
        )

    async def process_post(self, post: Post, force: bool = False) -> None:
        """
        Process a single post.
        """
        post_path = (
            self.root
            / "build"
            / self.path
            / post.path.relative_to(self.root / "posts").with_suffix(self.extension)
        )
        last_update = post_path.stat().st_mtime if post_path.exists() else None

        # create directories if needed
        post_directory = post_path.parent
        if not post_directory.exists():
            post_directory.mkdir(parents=True, exist_ok=True)

        if last_update and post.path.stat().st_mtime < last_update and not force:
            _logger.debug("Post %s is up-to-date, nothing to do", post_path)
            return

        template = self.env.get_template(
            f"{self.template_base}{post.type}{self.extension}",
        )
        content = template.render(
            config=self.config,
            post=post,
            home=self.home,
            render=self.render,
            __version__=__version__,
            sorted=sorted,
        )

        _logger.info("Creating %s post", self.label)
        with open(post_path, "w", encoding="utf-8") as output:
            output.write(content)

        for enclosure in post.enclosures:
            _logger.info("Copying enclosure %s", enclosure.path)
            target = post_directory / enclosure.path.relative_to(post.path.parent)
            target.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(enclosure.path, target)

    async def process_site(self, force: bool = False) -> None:
        """
        Process the entire site.
        """
        posts = get_posts(self.root, self.config)

        # build index and feed
        for asset in self.site_templates:
            path = self.root / "build" / self.path / asset
            template_name = f"{self.template_base}{asset}"
            self._build_index(path, template_name, posts, force)

        # template for groups (tags and categories)
        template_name = f"{self.template_base}group{self.extension}"

        # group posts by tag
        tags = defaultdict(list)
        for post in posts:
            for tag in sorted(post.tags):
                tags[tag].append(post)
        for tag, tag_posts in tags.items():
            path = self.root / "build" / self.path / "tags" / (tag + self.extension)
            self._build_index(path, template_name, tag_posts, force, title=tag)

        # group tags by categories
        categories = defaultdict(list)
        for post in posts:
            for category in sorted(post.categories):
                categories[category].append(post)
        for category, category_posts in categories.items():
            path = (
                self.root
                / "build"
                / self.path
                / "categories"
                / (category + self.extension)
            )
            title = self.config.categories[category].label
            subtitle = self.config.categories[category].description
            self._build_index(
                path,
                template_name,
                category_posts,
                force,
                title=title,
                subtitle=subtitle,
            )

    def _build_index(
        self,
        path: Path,
        template_name: str,
        posts: List[Post],
        force: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Build an index file from a list of posts.
        """
        last_update = path.stat().st_mtime if path.exists() else None

        if (
            last_update
            and all(post.path.stat().st_mtime < last_update for post in posts)
            and not force
        ):
            _logger.debug("File %s is up-to-date, nothing to do", path)
            return

        template = self.env.get_template(template_name)
        content = template.render(
            config=self.config,
            posts=posts,
            home=self.home,
            __version__=__version__,
            render=self.render,
            sorted=sorted,
            **kwargs,
        )

        _logger.info("Creating %s", path)
        with open(path, "w", encoding="utf-8") as output:
            output.write(content)


def get_builders(
    root: Path,
    config: Config,
) -> Dict[str, Builder]:
    """
    Return all the builders.
    """
    classes = {
        entry_point.name: entry_point.load()
        for entry_point in iter_entry_points("nefelibata.builder")
    }

    builders = {}
    for builder_name, builder_config in config.builders.items():
        class_ = classes[builder_config.plugin]

        builders[builder_name] = class_(root, config, **builder_config.dict())

    return builders



"""
Tests for ``nefelibata.builders.base``.
"""
from pathlib import Path
from typing import Type

from pytest_mock import MockerFixture
from yarl import URL

from nefelibata.builders.base import Builder, get_builders
from nefelibata.config import BuilderModel, Config
from nefelibata.post import Post

from ..conftest import MockEntryPoint


def make_dummy_builder(class_name: str) -> Type[Builder]:
    """
    Make a dummy ``Builder`` derived class.
    """
    return type(class_name, (Builder,), {})


def test_get_builders(
    mocker: MockerFixture,
    make_entry_point: Type[MockEntryPoint],
    root: Path,
    config: Config,
) -> None:
    """
    Test ``get_builders``.
    """
    DummyBuilder = make_dummy_builder("DummyBuilder")
    entry_points = [
        make_entry_point("builder", DummyBuilder),
    ]
    mocker.patch(
        "nefelibata.builders.base.iter_entry_points",
        return_value=entry_points,
    )

    config.builders = {
        "builder": BuilderModel(
            **{
                "plugin": "builder",
                "announce-on": [],
                "publish-to": [],
                "home": "https://example.com/",
                "path": "generic",
            }
        ),
    }
    builders = get_builders(root, config)
    assert len(builders) == 1
    assert isinstance(builders["builder"], DummyBuilder)


def test_builder_render(root: Path, config: Config) -> None:
    """
    Test the ``render`` method in ``Builder``.
    """
    builder = Builder(root, config, "https://example.com/")
    assert builder.render("test") == "test"


def test_builder_absolute_url(root: Path, config: Config, post: Post) -> None:
    """
    Test the ``absolute_url`` method in ``Builder``.
    """
    builder = Builder(root, config, "https://example.com/")
    builder.extension = ".html"
    assert builder.absolute_url(post) == URL("https://example.com/first/index.html")
